/**************************************************************************/
/*  Klavaro - a flexible touch typing tutor                               */
/*  Copyright (C) 2005 - 2008  Felipe Castro                              */
/*                                                                        */
/*  This program is free software, licensed under the terms of the GNU    */
/*  General Public License as published by the Free Software Foundation,  */
/*  which is also included in this package, in a file named "COPYING".    */
/**************************************************************************/

/*
 * Set of functions to deal with internationalization (translation).
 */
#include <stdio.h>
#include <string.h>
#include <glib.h>
#include <glib/gstdio.h>
#include <gtk/gtk.h>

#include "support.h"
#include "interface.h"
#include "main.h"
#include "callbacks.h"
#include "keyboard.h"
#include "velocity.h"
#include "fluidness.h"
#include "translation.h"

/**********************************************************************
 * Variables
 */
extern GtkWidget *window_main_;
extern GtkWidget *window_top10_;

Lang_Name_Code *lang;
gint lang_num = 0;

/**********************************************************************
 * Set the code of the current language defined by environmental 
 * variables and apply it for the GTK interface
 */
void
trans_set_language_env (gchar * code)
{
	if (g_setenv ("LANGUAGE", code, 1) == FALSE)
		if (g_setenv ("LANG", code, 1) == FALSE)
			g_warning ("no language environment variable could be set.");
	main_preferences_set_string ("interface", "language", code);
	gtk_set_locale ();
}

/**********************************************************************
 * Initialize 'lang' and 'code' accordingly to 'language_set',
 * which was defined in "languages.h".
 */
void
trans_init_lang_name_code ()
{
	gint i;
	gchar delimiter[] = "\n";
	gchar tmp_str[LANG_NAME_MAX_LEN];
	gchar **temp_lang = NULL;
	const gchar *languages_set = LANG_SET;

	/* 
	 * Get the number of configured languages
	 */
	temp_lang = g_strsplit (languages_set, delimiter, -1);
	i = 0;
	while (temp_lang[i] != NULL)
		i++;
	g_assert (i > 0);
	lang = g_new (Lang_Name_Code, i);
	lang_num = i;

	delimiter[1] = '(';
	for (i = 0; i < lang_num; i++)
	{
		/* Initialize 'lang'
		 */
		strncpy (tmp_str, temp_lang[i], LANG_NAME_MAX_LEN - 1);
		tmp_str[LANG_NAME_MAX_LEN] = '\0';
		g_strdelimit (tmp_str, delimiter, '\0');
		strcpy (lang[i].name, g_strstrip (tmp_str));

		/* Initialize 'code'
		 */
		strcpy (tmp_str, strchr (temp_lang[i], '('));
		*(strrchr (tmp_str, ')')) = '\0';
		g_assert (strlen (tmp_str + 1) < 6);
		strcpy (lang[i].code, tmp_str + 1);

		/* Initialize 'cd'
		 */
		if (lang[i].code[0] == 'C')
		{
			lang[i].cd[0] = 'e';
			lang[i].cd[1] = 'n';
		}
		else
		{
			lang[i].cd[0] = lang[i].code[0];
			lang[i].cd[1] = lang[i].code[1];
		}
		lang[i].cd[2] = '\0';
	}
	g_strfreev (temp_lang);
}

gchar *
trans_get_code (gint i)
{
	if (i >= lang_num || i < 0)
		return (NULL);
	return (lang[i].cd);
}

gboolean
trans_lang_is_available (gchar * test)
{
	gint i;

	for (i = 0; i < lang_num; i++)
		if (g_str_equal (lang[i].code, test))
			break;
	return (i == lang_num ? FALSE : TRUE);
}

/**********************************************************************
 * Define if we may put a stop mark at the end of "phrases".
 */
gboolean
trans_lang_has_stopmark ()
{
	gboolean stopmark;
	gchar *hlp;

	hlp = main_preferences_get_string ("interface", "language");
	stopmark = g_str_has_prefix (hlp, "ur") ||
		   g_str_has_prefix (hlp, "bn");
	g_free (hlp);

	return (!stopmark);
}

/**********************************************************************
 * Private auxiliar function
 */
static gboolean
trans_lang_get_similar (gchar * test)
{
	gint i;
	gchar aux_code_2[3];

	if (g_str_equal (test, "C"))
		return TRUE;

	strncpy (aux_code_2, test, 2);
	aux_code_2[2] = '\0';

	for (i = 0; i < lang_num; i++)
	{
		if (strstr (lang[i].code, aux_code_2))
		{
			g_free (test);
			test = g_strdup (lang[i].code);
			break;
		}
	}
	if (i == lang_num && g_str_has_prefix (test, "en"))
	{
		g_free (test);
		test = g_strdup ("C");
		return (TRUE);
	}
	return (i == lang_num ? FALSE : TRUE);
}

/**********************************************************************
 * Get the current locale and change it if necessary
 */
void
trans_init_language_env ()
{
	gchar *tmp_code = NULL;
	gboolean lang_ok = TRUE;
	gint i;

	trans_init_lang_name_code ();

	/*
	 * If the language is already set in preferences, just use it
	 */
	if (main_preferences_exist ("interface", "language"))
		tmp_code = main_preferences_get_string ("interface", "language");
	else
	{
		/* 
		 * Read the current locale
		 */
#ifdef G_OS_UNIX
		i = 0;
		lang_ok = FALSE;
		while ((tmp_code = g_strdup (g_get_language_names ()[i])))
		{
			if (tmp_code[0] == 'C')
			{
				lang_ok = (i == 0 ? TRUE : FALSE);
				break;
			}
			lang_ok = trans_lang_is_available (tmp_code);
			if (lang_ok == TRUE)
				break;
			g_free (tmp_code);
			lang_ok = FALSE;
			i++;
		}
		if (lang_ok == FALSE)
		{
			i = 0;
			while ((tmp_code = g_strdup (g_get_language_names ()[i])))
			{
				if (tmp_code[0] == 'C')
				{
					lang_ok = (i == 0 ? TRUE : FALSE);
					break;
				}
				lang_ok = trans_lang_get_similar (tmp_code);
				if (lang_ok == TRUE)
					break;
				g_free (tmp_code);
				lang_ok = FALSE;
				i++;
			}
		}
#else
		tmp_code = g_win32_getlocale ();
		lang_ok = trans_lang_is_available (tmp_code);
		if (lang_ok == FALSE)
			lang_ok = trans_lang_get_similar (tmp_code);
#endif
	}
	if (tmp_code == NULL)
		tmp_code = g_strdup ("xx");
	/*
	 * If even a similar is not available, then use the neutral international one
	 */
	if (lang_ok == FALSE)
	{
		g_free (tmp_code);
		tmp_code = g_strdup ("eo");
		g_message ("as your prefered locales aren't available,");
		g_message ("we are using this neutral one: \"%s\"", tmp_code);
		g_message ("if you don't agree, try to change it at the main menu... ;-)");
	}

	trans_set_language_env (tmp_code);
	g_free (tmp_code);
}

/**********************************************************************
 * Inserts the list of available languages in the 'combo_language'.
 */
void
trans_set_combo_language ()
{
	gint i;
	gint i_env;
	gchar *tmp_code;
	GList *languages = NULL;
	GtkWidget *wg;

	tmp_code = main_preferences_get_string ("interface", "language");
	if (strcmp (tmp_code, "eo_XX") == 0)
	{
		main_preferences_set_string ("interface", "language", "eo");
		g_free (tmp_code);
		tmp_code = g_strdup ("eo");
	}

	for (i = 0, i_env = -1; i < lang_num; i++)
	{
		if (strcmp (lang[i].code, tmp_code) != 0)
			languages = g_list_insert (languages, lang[i].name, -1);
		else
			i_env = i;
	}

	if (i_env == -1)
		g_error ("set_combo_language() ==> the locale \"%s\" is not available!", tmp_code);
	else
		languages = g_list_insert (languages, lang[i_env].name, 0);

	wg = lookup_widget (window_main_, "combo_language");
	gtk_combo_set_popdown_strings (GTK_COMBO (wg), languages);

	g_free (tmp_code);
	g_list_free (languages);
}

/**********************************************************************
 * Get the current language defined by the preference's key and set in
 * the 'combo_language'.
 */
gchar *
trans_get_current_language ()
{
	gchar *tmp_code;
	gint i;

	tmp_code = main_preferences_get_string ("interface", "language");
	for (i = 0; i < lang_num; i++)
		if (strcmp (lang[i].code, tmp_code) == 0)
		{
			g_free (tmp_code);
			return (lang[i].name);
		}
	g_free (tmp_code);
	return ("??");
}

/**********************************************************************
 * Update the current language used accordingly to that selected in the
 * 'combo_language'
 */
void
trans_change_language ()
{
	gint i;
	gchar *selection = NULL;
	GtkWidget *wg;

	wg = lookup_widget (window_main_, "entry_language");
	selection = gtk_editable_get_chars (GTK_EDITABLE (wg), 0, -1);

	for (i = 0; i < lang_num; i++)
		if (strcmp (lang[i].name, selection) == 0)
			break;

	if (i == lang_num)
	{
		g_warning ("change_language() --> couldn't find the language: %s", selection);
		g_free (selection);
		return;
	}
	g_free (selection);

	trans_set_language_env (lang[i].code);
	g_message ("selected language code --> %s", lang[i].code);

	/* Reset everything up */
	callbacks_shield_set (TRUE);
	gtk_widget_destroy (window_main_);
	window_main_ = create_window_main ();
	wg = lookup_widget (window_main_, "entry_keyboard");
	gtk_entry_set_text (GTK_ENTRY (wg), keyb_get_name ());
	trans_set_combo_language ();
	main_set_auto_state ();
	gtk_widget_show (window_main_);
	callbacks_shield_set (FALSE);

	velo_reset_dict ();
	fluid_reset_paragraph ();

	if (window_top10_ != NULL)
	{
		gtk_widget_destroy (window_top10_);
		window_top10_ = NULL;
	}
}

/**********************************************************************
 * Find a file whose language prefix is similar to the current one
 */
FILE *
trans_lang_get_similar_file (const gchar * file_end)
{
	gint i;
	gchar *tmp_code;
	gchar *tmp_path = NULL;
	FILE *fh = NULL;

	tmp_code = main_preferences_get_string ("interface", "language");
	for (i = 0; i < lang_num && fh == NULL; i++)
	{
		if (strcmp (lang[i].code, tmp_code) == 0)
			continue;
		if (lang[i].code[0] == tmp_code[0] && lang[i].code[1] == tmp_code[1])
		{
			g_free (tmp_path);
			tmp_path =
				g_strconcat (main_get_data_path (), lang[i].code, file_end, NULL);
			fh = (FILE *) g_fopen (tmp_path, "r");
		}
	}
	g_free (tmp_code);
	g_free (tmp_path);
	i--;
	if (fh)
		g_message ("applying similar file: %s%s", lang[i].code, file_end);
	return (fh);
}

/**********************************************************************
 * Find a file whose language prefix is similar to the current one, returnig
 * its name
 */
gchar *
trans_lang_get_similar_file_name (const gchar * file_end)
{
	gint i;
	gchar *tmp_code;
	gchar *tmp_path = NULL;

	tmp_code = main_preferences_get_string ("interface", "language");
	for (i = 0; i < lang_num; i++)
	{
		if (strcmp (lang[i].code, tmp_code) == 0)
			continue;
		if (lang[i].code[0] == tmp_code[0] && lang[i].code[1] == tmp_code[1])
		{
			g_free (tmp_path);
			tmp_path =
				g_strconcat (main_get_data_path (), lang[i].code, file_end, NULL);
			if (g_file_test (tmp_path, G_FILE_TEST_IS_REGULAR))
				break;
		}
	}
	if (tmp_path == NULL)
		tmp_path = g_strconcat (main_get_data_path (), "C", file_end, NULL);
	g_free (tmp_code);
	return (tmp_path);
}

/**********************************************************************
 * Reads the text of a file to be presented at some dialog,
 * accordingly to the environment language.
 * 'text': new buffer where the text is copied into.
 * 'file_end': how the file name ends
 */
#define AUTHOR "Felipe E. F. de Castro"
#define EMAIL "fefcas@gmail.com"
#define HOMEPAGE "http://klavaro.sf.net"
gchar *
trans_read_text (const gchar * file_end)
{
	gchar *buf;
	gchar *tmp_name = NULL;
	gchar *tmp_path = NULL;
	gchar *tmp_code;
	gint i, j, tmpc;
	FILE *fh;

	gchar *help1 = _(
"This is the main menu of the application and it is very straightforward. "
"Just select the type of exercise you wish to take. "
"Do not worry about clicking any button on the screen. "
"The buttons will never hurt you!");
	gchar *help2 = _(
"Also the controls have tips which will appear if you hover over them with your mouse. "
"These tips can come in very handy. "
"To see how these tips work, position your mouse over the button below, "
"and without clicking, wait for the tip to appear...");

	gchar *about1 = _("Yet another touch typing tutor");
	gchar *about2 = _("Version:");
	gchar *about3 = _("by:");
	gchar *about4 = _("E-mail:");
	gchar *about5 = _("Home-page:");
	gchar *about6 = _("Remember always: someone loves you!");

	gchar *intro01 = _(
"Correct positioning of the hands and fingers is very important to efficient typing. "
"You will learn faster and type better if you follow these recomendations.");
	gchar *intro02 = _("To position the hands:");
	gchar *intro03 = _(
"1)  The index-finger tips rest over each of the two keys which have a small raised mark, "
"in the center of the keyboard. Most keyboards are of the 'qwerty' type, "
"where these two keys are the 'F' and the 'J' keys. On the other hand, "
"in an American Dvorak keyboard, they are the 'U' and the 'H' keys.");
	gchar *intro04 = _(
"obs.: these marks function as 'tactile hooks' for your fingers to remain at the correct position; "
"this way, with a little experience, you will not need to look at the keyboard to see if your "
"fingers are properly positioned.");
	gchar *intro05 = _(
"2)  The outside edges of your thumbs rest over the space bar.");
	gchar *intro06 = _(
"3)  The tips of the other fingers lie naturally beside the index ones, "
"over the keys on the same row of the keyboard.");
	gchar *intro07 = _(
"4)  The part of the hands closest to the wrist (the base) rest over the table, "
"outside the keyboard. Without this kind of support the arms would quickly tire.");
	gchar *intro08 = _(
"This is referred to as the home position for the hands. "
"From it the fingers move all over the keyboard, "
"reaching all the keys as naturally and quickly as possible. "
"To reach this goal one uses a specific relation between each key and finger. "
"This relation will be learned gradually as you complete the basic course.");
	gchar *intro09 = _(
"When learning the relation between fingers and keys, "
"it is very important that you only move the finger which must press the key "
"and allow all other fingers to remain in the home position.");
	gchar *intro10 = _(
"After memorizing this relationship, you can relax the previous rule some, "
"so that you can attain greater speed while typing.");
	gchar *intro11 = _(
"You should be prepared to start training with the basic course. "
"It will take effort and patience to be successful as a typist. "
"We trust you have these and look forward to your success!");

	gchar *basic1 = _(
"The basic course focuses on having you read the characters presented to you on screen "
"and typing the corresponding keys. Remember to keep your hands correctly oriented "
"on the home row of the keyboard at all times (see introduction on main menu).");
	gchar *basic2 = _(
"The key set used in each series will be shown in the above message line. "
"The [Space], [Shift] and [Enter] keys may not show up there but are used very often.");
	gchar *basic3 = _(
"The message line below follows and echoes your key presses. "
"If required, it changes and displays instructions for actions required from you.");

	gchar *adapt1 = _(
"Here you may practice and improve your memorization of all keys. "
"There will be sentences presented with nonsense words which mix some numbers and symbols.");
	gchar *adapt2 = _(
"In order to keep the lesson contents language and keyboard independent, "
"accented letter combinations will probably not appear. For real word sentences, "
"please use the fourth option of the main menu (about fluidness).");
	gchar *adapt3 = _(
"After each exercise there will be a brief statistics panel "
"reviewing your performance along with some relevant comments.");

	gchar *velo1 = _(
"This exercise is very similar to the second one, for adaptability. "
"The difference is that here you'll type real words.");
	gchar *velo2 = _(
"The default language is the actual one of the interface. "
"But you may select any other texts with words you would like to use. "
"Press the 'Other' option above and add files containing those texts.");
	gchar *velo3 = _(
"With this exercise the focus is on speed. "
"So, you are supposed to type really fast and I will only flatter you when you deserve it!");

	gchar *fluid1 = _(
"We will now use complete sentences and paragraphs which make logical sense. "
"This may distract you while you type if you try to understand what you are entering. "
"The previous exercises were aimed at getting you to type without interpreting and analyzing the content.");
	gchar *fluid2 = _(
"We do not mean to imply that the typists must behave like a robot, without understanding what they type. "
"We do aim to develop the skill of typing, making it an automatic reflex akin to the acts of walking, talking, etc. "
"After reaching this goal, the act of typing will become automatic and require little concentration. "
"Then you will be able to pay attention to the real meaning of the text.");
	gchar *fluid3 = _(
"These exercises are longer. Each exercise consists of three paragraphs and "
"the emphasis is placed on correctness and rhythm, with a minimun speed requirement. "
"Here you will be required to use the backspace key to correct any mistakes. "
"In other words, only input without error will be accepted.");

	if (g_str_equal (file_end, "_help.txt"))
		return (g_strdup_printf ("%s\n%s", help1, help2));

	if (g_str_equal (file_end, "_about.txt"))
		return (g_strdup_printf ("%s\n%s " PACKAGE_VERSION "\n%s " AUTHOR "\n%s " EMAIL "\n%s " HOMEPAGE "\n%s", about1, about2, about3, about4, about5, about6));

	if (g_str_equal (file_end, "_intro.txt"))
		return (g_strdup_printf ("%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s", intro01, intro02, intro03, intro04, intro05, intro06, intro07, intro08, intro09, intro10, intro11));

	if (g_str_equal (file_end, "_basic_intro.txt"))
		return (g_strdup_printf ("%s\n%s\n%s", basic1, basic2, basic3));

	if (g_str_equal (file_end, "_adapt_intro.txt"))
		return (g_strdup_printf ("%s\n%s\n%s", adapt1, adapt2, adapt3));

	if (g_str_equal (file_end, "_velo_intro.txt"))
		return (g_strdup_printf ("%s\n%s\n%s", velo1, velo2, velo3));

	if (g_str_equal (file_end, "_fluid_intro.txt"))
		return (g_strdup_printf ("%s\n%s\n%s", fluid1, fluid2, fluid3));

	/* Use text files
	 */
	tmp_code = main_preferences_get_string ("interface", "language");
	tmp_name = g_strconcat (tmp_code, file_end, NULL);

	/* Try at HOME
	 */
	tmp_path = g_strconcat (main_get_user_dir (), tmp_name, NULL);
	fh = (FILE *) g_fopen (tmp_path, "r");

	/* Try at PACKAGE_DATA
	 */
	if (fh == NULL)
	{
		g_free (tmp_path);
		tmp_path = g_strconcat (main_get_data_path (), tmp_name, NULL);
		fh = (FILE *) g_fopen (tmp_path, "r");
	}

	/*
	 * Try other "flavors" of the same language
	 */
	if (fh == NULL && strlen (tmp_code) > 1)
		fh = trans_lang_get_similar_file (file_end);

	/*
	 * Default to C
	 */
	if (fh == NULL && strcmp (tmp_code, "C") != 0)
	{
		g_message
			("trans_read_text() --> couldn't open the data file: %s\n"
			 " So, we have to apply the default one: C%s", tmp_name, file_end);
		main_preferences_set_string ("interface", "language", "C");
		buf = trans_read_text (file_end);
		main_preferences_set_string ("interface", "language", tmp_code);
		g_free (tmp_code);
		g_free (tmp_path);
		g_free (tmp_name);
		return buf;
	}

	if (fh == NULL)
		g_error ("trans_read_text() --> couldn't open the data file:\n %s", tmp_name);

	g_free (tmp_code);
	g_free (tmp_path);
	g_free (tmp_name);

	/*
	 * Process the file
	 */
	gsize bufsize = 16;
	gsize pos = 0;
	buf = g_malloc (bufsize);
	while (1)
	{
		gsize max = bufsize - pos - 1; // -1 for terminating zero
		gsize ct = fread (buf + pos, 1, max, fh);
		if (ct == 0)
			break;
		pos += ct;
		if (ct == max)
		{
			bufsize *= 2;
			buf = g_realloc (buf, bufsize);
		}
	}
	buf[pos] = 0;
	return buf;
}
